"use strict";Object.defineProperty(exports, "__esModule", {value: true}); function _nullishCoalesce(lhs, rhsFn) { if (lhs != null) { return lhs; } else { return rhsFn(); } } function _optionalChain(ops) { let lastAccessLHS = undefined; let value = ops[0]; let i = 1; while (i < ops.length) { const op = ops[i]; const fn = ops[i + 1]; i += 2; if ((op === 'optionalAccess' || op === 'optionalCall') && value == null) { return undefined; } if (op === 'access' || op === 'optionalAccess') { lastAccessLHS = value; value = fn(value); } else if (op === 'call' || op === 'optionalCall') { value = fn((...args) => value.call(lastAccessLHS, ...args)); lastAccessLHS = undefined; } } return value; }var _biggystring = require('biggystring');









var _cleaners = require('cleaners');











var _types = require('edge-core-js/types');

var _biggystringplus = require('../../util/biggystringplus');




var _swapHelpers = require('../../util/swapHelpers');








var _utils = require('../../util/utils');


const swapInfo = {
  pluginId: 'rango',
  isDex: true,
  displayName: 'Rango Exchange',
  supportEmail: 'support@edge.app'
}

const EXPIRATION_MS = 1000 * 60
const EXCHANGE_INFO_UPDATE_FREQ_MS = 60000

const MAINNET_CODE_TRANSCRIPTION = {
  arbitrum: 'ARBITRUM',
  // axelar: 'AXELAR',
  avalanche: 'AVAX_CCHAIN',
  base: 'BASE',
  binancesmartchain: 'BSC',
  // bitcoin: 'BTC',
  // celo: 'CELO',
  // cosmoshub: 'COSMOS',
  // dash: 'DASH',
  // dogecoin: 'DOGE',
  ethereum: 'ETH',
  fantom: 'FANTOM',
  // injective: 'INJECTIVE',
  // litecoin: 'LTC',
  // maya: 'MAYA',
  // moonbeam: 'MOONBEAM',
  // moonriver: 'MOONRIVER',
  // okexchain: 'OKC',
  optimism: 'OPTIMISM',
  // osmosis: 'OSMOSIS',
  polygon: 'POLYGON',
  // thorchainrune: 'THOR',
  // solana: 'SOLANA',
  // tron: 'TRON',
  zksync: 'ZKSYNC'
}

const RANGO_SERVERS_DEFAULT = ['https://api.rango.exchange']

const PARENT_TOKEN_CONTRACT_ADDRESS = '0x0'

const DEFAULT_SLIPPAGE = '5.0'







function assetToString(asset) {
  if (
    !(asset.address == null) &&
    asset.address !== PARENT_TOKEN_CONTRACT_ADDRESS
  )
    return `${asset.blockchain}.${asset.symbol}--${asset.address}`
  else return `${asset.blockchain}.${asset.symbol}`
}

const asInitOptions = _cleaners.asObject.call(void 0, {
  appId: _cleaners.asOptional.call(void 0, _cleaners.asString, 'edge'),
  rangoApiKey: _cleaners.asString,
  referrerAddress: _cleaners.asOptional.call(void 0, _cleaners.asString),
  referrerFee: _cleaners.asOptional.call(void 0, _cleaners.asString)
})

const asExchangeInfo = _cleaners.asObject.call(void 0, {
  swap: _cleaners.asObject.call(void 0, {
    plugins: _cleaners.asObject.call(void 0, {
      rango: _cleaners.asOptional.call(void 0, 
        _cleaners.asObject.call(void 0, {
          rangoServers: _cleaners.asOptional.call(void 0, _cleaners.asArray.call(void 0, _cleaners.asString))
        })
      )
    })
  })
})

const asCompactToken = _cleaners.asObject.call(void 0, {
  // blockchain
  b: _cleaners.asString,
  // address
  a: _cleaners.asOptional.call(void 0, _cleaners.asString),
  // symbol
  s: _cleaners.asString
})

const asCompactMetaResponse = _cleaners.asObject.call(void 0, {
  tokens: _cleaners.asArray.call(void 0, asCompactToken)
})

const asToken = _cleaners.asObject.call(void 0, {
  blockchain: _cleaners.asString,
  address: _cleaners.asEither.call(void 0, _cleaners.asString, _cleaners.asNull),
  symbol: _cleaners.asString
})

const asSwapperMeta = _cleaners.asObject.call(void 0, {
  id: _cleaners.asString,
  title: _cleaners.asString
})

const asSwapPath = _cleaners.asObject.call(void 0, {
  from: asToken,
  to: asToken,
  swapper: asSwapperMeta,
  expectedOutput: _cleaners.asString
})

const asSwapFee = _cleaners.asObject.call(void 0, {
  name: _cleaners.asString,
  token: asToken,
  expenseType: _cleaners.asValue.call(void 0, 
    'FROM_SOURCE_WALLET',
    'DECREASE_FROM_OUTPUT',
    'FROM_DESTINATION_WALLET'
  ),
  amount: _cleaners.asString
})

const asAmountRestriction = _cleaners.asObject.call(void 0, {
  min: _cleaners.asString,
  max: _cleaners.asString,
  type: _cleaners.asString // "EXCLUSIVE"
})

const asSwapSimulationResult = _cleaners.asObject.call(void 0, {
  from: asToken,
  to: asToken,
  amountRestriction: _cleaners.asOptional.call(void 0, asAmountRestriction),
  outputAmount: _cleaners.asString,
  outputAmountMin: _cleaners.asString,
  outputAmountUsd: _cleaners.asEither.call(void 0, _cleaners.asNumber, _cleaners.asNull),
  swapper: asSwapperMeta,
  path: _cleaners.asEither.call(void 0, _cleaners.asArray.call(void 0, asSwapPath), _cleaners.asNull),
  fee: _cleaners.asArray.call(void 0, asSwapFee),
  feeUsd: _cleaners.asEither.call(void 0, _cleaners.asNumber, _cleaners.asNull),
  estimatedTimeInSeconds: _cleaners.asNumber
})

const asRoutingResultType = _cleaners.asValue.call(void 0, 
  'OK',
  'HIGH_IMPACT',
  'NO_ROUTE',
  'INPUT_LIMIT_ISSUE'
)

const asEvmTransaction = _cleaners.asObject.call(void 0, {
  type: _cleaners.asValue.call(void 0, 'EVM'),
  from: _cleaners.asEither.call(void 0, _cleaners.asString, _cleaners.asNull),
  approveTo: _cleaners.asEither.call(void 0, _cleaners.asString, _cleaners.asNull),
  approveData: _cleaners.asEither.call(void 0, _cleaners.asString, _cleaners.asNull),
  txTo: _cleaners.asString,
  txData: _cleaners.asEither.call(void 0, _cleaners.asString, _cleaners.asNull),
  value: _cleaners.asEither.call(void 0, _cleaners.asString, _cleaners.asNull),
  gasLimit: _cleaners.asEither.call(void 0, _cleaners.asString, _cleaners.asNull),
  gasPrice: _cleaners.asEither.call(void 0, _cleaners.asString, _cleaners.asNull),
  maxPriorityFeePerGas: _cleaners.asEither.call(void 0, _cleaners.asString, _cleaners.asNull),
  maxFeePerGas: _cleaners.asEither.call(void 0, _cleaners.asString, _cleaners.asNull)
})

const asSwapResponse = _cleaners.asObject.call(void 0, {
  resultType: asRoutingResultType,
  route: _cleaners.asEither.call(void 0, asSwapSimulationResult, _cleaners.asNull),
  error: _cleaners.asEither.call(void 0, _cleaners.asString, _cleaners.asNull),
  tx: _cleaners.asEither.call(void 0, asEvmTransaction, _cleaners.asNull)
})





const rangoTokens

 = {}

let exchangeInfo
let exchangeInfoLastUpdate = 0

 function makeRangoPlugin(opts) {
  const { io, log } = opts
  const { fetchCors = io.fetch } = io
  const { appId, rangoApiKey, referrerAddress, referrerFee } = asInitOptions(
    opts.initOptions
  )

  const headers = {
    'Content-Type': 'application/json'
  }

  let rangoServers = RANGO_SERVERS_DEFAULT

  let params = _utils.makeQueryParams.call(void 0, {
    apiKey: rangoApiKey
  })

  Object.values(MAINNET_CODE_TRANSCRIPTION).forEach(
    blockchain => (params += `&blockchains=${blockchain}`)
  )

  const metaRequest = _utils.fetchWaterfall.call(void 0, 
    fetchCors,
    rangoServers,
    `meta/compact?${params}`,
    {
      headers
    }
  )

  const fetchSwapQuoteInner = async (
    request
  ) => {
    const {
      fromTokenId,
      toTokenId,
      nativeAmount,
      fromWallet,
      toWallet,
      quoteFor
    } = request
    if (quoteFor !== 'from') {
      throw new (0, _types.SwapCurrencyError)(swapInfo, request)
    }

    if (Object.keys(rangoTokens).length === 0) {
      await metaRequest
        .then(async metaResponse => {
          if (metaResponse.ok) {
            return await metaResponse.json()
          } else {
            const text = await metaResponse.text()
            throw new Error(text)
          }
        })
        .then(meta => {
          const { tokens } = asCompactMetaResponse(meta)
          tokens.forEach(token => {
            const tokenBlockchain = token.b
            const tokenAddress = token.a
            const tokenSymbol = token.s
            if (rangoTokens[tokenBlockchain] === undefined) {
              rangoTokens[tokenBlockchain] = {}
            }
            rangoTokens[tokenBlockchain][
              _nullishCoalesce(tokenAddress, () => ( PARENT_TOKEN_CONTRACT_ADDRESS))
            ] = tokenSymbol
          })
        })
        .catch(e => {
          throw new Error(`Error fetching Rango meta ${String(e)}`)
        })
    }

    const fromToken =
      fromTokenId != null
        ? fromWallet.currencyConfig.allTokens[fromTokenId]
        : undefined
    let fromContractAddress
    if (fromTokenId === null) {
      fromContractAddress = PARENT_TOKEN_CONTRACT_ADDRESS
    } else {
      fromContractAddress = _optionalChain([fromToken, 'optionalAccess', _ => _.networkLocation, 'optionalAccess', _2 => _2.contractAddress])
    }

    const toToken =
      toTokenId != null
        ? toWallet.currencyConfig.allTokens[toTokenId]
        : undefined
    let toContractAddress
    if (toTokenId === null) {
      toContractAddress = PARENT_TOKEN_CONTRACT_ADDRESS
    } else {
      toContractAddress = _optionalChain([toToken, 'optionalAccess', _3 => _3.networkLocation, 'optionalAccess', _4 => _4.contractAddress])
    }

    if (fromContractAddress == null || toContractAddress == null) {
      throw new (0, _types.SwapCurrencyError)(swapInfo, request)
    }

    // Do not support transfer between same assets
    if (
      fromWallet.currencyInfo.pluginId === toWallet.currencyInfo.pluginId &&
      request.fromCurrencyCode === request.toCurrencyCode
    ) {
      throw new (0, _types.SwapCurrencyError)(swapInfo, request)
    }

    const fromAddress = await _utils.getAddress.call(void 0, fromWallet)
    const toAddress = await _utils.getAddress.call(void 0, toWallet)

    const fromMainnetCode =
      MAINNET_CODE_TRANSCRIPTION[fromWallet.currencyInfo.pluginId]
    const toMainnetCode =
      MAINNET_CODE_TRANSCRIPTION[toWallet.currencyInfo.pluginId]

    if (fromMainnetCode == null || toMainnetCode == null) {
      throw new (0, _types.SwapCurrencyError)(swapInfo, request)
    }

    const fromSymbol =
      _optionalChain([rangoTokens, 'access', _5 => _5[fromMainnetCode], 'optionalAccess', _6 => _6[fromContractAddress.toLowerCase()]])
    const toSymbol =
      _optionalChain([rangoTokens, 'access', _7 => _7[toMainnetCode], 'optionalAccess', _8 => _8[toContractAddress.toLowerCase()]])

    if (fromSymbol == null || toSymbol == null) {
      throw new (0, _types.SwapCurrencyError)(swapInfo, request)
    }

    const now = Date.now()
    if (
      now - exchangeInfoLastUpdate > EXCHANGE_INFO_UPDATE_FREQ_MS ||
      exchangeInfo == null
    ) {
      try {
        const exchangeInfoResponse = await _utils.promiseWithTimeout.call(void 0, 
          _utils.fetchInfo.call(void 0, fetchCors, `v1/exchangeInfo/${appId}`)
        )

        if (exchangeInfoResponse.ok === true) {
          exchangeInfo = asExchangeInfo(await exchangeInfoResponse.json())
          exchangeInfoLastUpdate = now
        } else {
          // Error is ok. We just use defaults
          const text = await exchangeInfoResponse.text()
          log.warn(
            `Error getting info server exchangeInfo. Using defaults... Error: ${text}`
          )
        }
      } catch (e) {
        log.warn(
          'Error getting info server exchangeInfo. Using defaults...',
          e.message
        )
      }
    }

    if (exchangeInfo != null) {
      const { rango } = exchangeInfo.swap.plugins
      rangoServers = _nullishCoalesce(_optionalChain([rango, 'optionalAccess', _9 => _9.rangoServers]), () => ( rangoServers))
    }

    let referrer

    if (
      referrerAddress != null &&
      referrerAddress !== '' &&
      referrerFee != null &&
      referrerFee !== ''
    ) {
      referrer = { referrerAddress, referrerFee }
    }

    const swapParameters = {
      apiKey: rangoApiKey,
      from: assetToString({
        blockchain: fromMainnetCode,
        address: fromContractAddress,
        symbol: fromSymbol
      }),
      to: assetToString({
        blockchain: toMainnetCode,
        address: toContractAddress,
        symbol: toSymbol
      }),
      fromAddress: fromAddress,
      toAddress: toAddress,
      amount: nativeAmount,
      disableEstimate: true,
      slippage: DEFAULT_SLIPPAGE,
      ...(referrer != null ? referrer : undefined)
    }

    const params = _utils.makeQueryParams.call(void 0, swapParameters)

    const swapResponse = await _utils.fetchWaterfall.call(void 0, 
      fetchCors,
      rangoServers,
      `basic/swap?${params}`,
      {
        headers
      }
    )

    if (!swapResponse.ok) {
      const responseText = await swapResponse.text()
      throw new Error(`Rango could not fetch quote: ${responseText}`)
    }

    const swap = asSwapResponse(await swapResponse.json())
    const { route, tx } = swap

    if (swap.resultType !== 'OK') {
      if (swap.resultType === 'INPUT_LIMIT_ISSUE') {
        const amountRestriction = _optionalChain([swap, 'access', _10 => _10.route, 'optionalAccess', _11 => _11.amountRestriction])
        if (amountRestriction == null) {
          throw new Error('Rango limit error without values')
        }
        const { min, max } = amountRestriction

        if (_biggystring.gte.call(void 0, nativeAmount, max)) {
          throw new (0, _types.SwapAboveLimitError)(swapInfo, max)
        } else if (_biggystring.lte.call(void 0, nativeAmount, min)) {
          throw new (0, _types.SwapBelowLimitError)(swapInfo, min)
        }
      }
      throw new Error(
        `Rango could not proceed with the exchange. : ${swap.resultType}`
      )
    }

    if (
      _optionalChain([route, 'optionalAccess', _12 => _12.path]) == null ||
      route.outputAmount === '' ||
      tx == null ||
      tx.txData == null
    ) {
      throw new Error('Rango could not proceed with the exchange')
    }

    const providers = route.path.map(p => p.swapper.title)

    let preTx
    if (tx.type === 'EVM' && tx.approveData != null && tx.approveTo != null) {
      const approvalData = tx.approveData.replace('0x', '')

      const spendInfo = {
        tokenId: null,
        memos: [{ type: 'hex', value: approvalData }],
        spendTargets: [
          {
            nativeAmount: '0',
            publicAddress: fromContractAddress
          }
        ],
        assetAction: {
          assetActionType: 'tokenApproval'
        },
        savedAction: {
          actionType: 'tokenApproval',
          tokenApproved: {
            pluginId: fromWallet.currencyInfo.pluginId,
            tokenId: fromTokenId,
            nativeAmount
          },
          tokenContractAddress: fromContractAddress,
          contractAddress: tx.approveTo
        }
      }
      preTx = await request.fromWallet.makeSpend(spendInfo)
    }

    const customNetworkFee = {
      gasLimit: tx.gasLimit != null ? _utils.hexToDecimal.call(void 0, tx.gasLimit) : undefined,
      gasPrice:
        tx.gasPrice != null ? _biggystringplus.div18.call(void 0, tx.gasPrice, '1000000000') : undefined,
      maxFeePerGas: _nullishCoalesce(tx.maxFeePerGas, () => ( undefined)),
      maxPriorityFeePerGas: _nullishCoalesce(tx.maxPriorityFeePerGas, () => ( undefined))
    }

    const networkFeeOption =
      customNetworkFee.gasLimit != null || customNetworkFee.gasPrice != null
        ? 'custom'
        : undefined

    const value = tx.txData.replace('0x', '')
    const spendInfo = {
      tokenId: request.fromTokenId,
      memos: [{ type: 'hex', value }],
      customNetworkFee,
      spendTargets: [
        {
          memo: tx.txData,
          nativeAmount: nativeAmount,
          publicAddress: tx.txTo
        }
      ],
      networkFeeOption,
      assetAction: {
        assetActionType: 'swap'
      },
      savedAction: {
        actionType: 'swap',
        swapInfo,
        isEstimate: true,
        toAsset: {
          pluginId: toWallet.currencyInfo.pluginId,
          tokenId: toTokenId,
          nativeAmount: route.outputAmount
        },
        fromAsset: {
          pluginId: fromWallet.currencyInfo.pluginId,
          tokenId: fromTokenId,
          nativeAmount: nativeAmount
        },
        payoutAddress: toAddress,
        payoutWalletId: toWallet.id,
        refundAddress: fromAddress
      }
    }

    const providersStr = _optionalChain([providers, 'optionalAccess', _13 => _13.join, 'call', _14 => _14(' -> ')])
    const metadataNotes = `DEX Providers: ${providersStr}`

    return {
      request,
      spendInfo,
      swapInfo,
      fromNativeAmount: nativeAmount,
      expirationDate: new Date(Date.now() + EXPIRATION_MS),
      preTx,
      metadataNotes
    }
  }

  const out = {
    swapInfo,

    async fetchSwapQuote(req) {
      const request = _utils.convertRequest.call(void 0, req)

      let newRequest = request
      if (request.quoteFor === 'max') {
        if (request.fromTokenId != null) {
          const maxAmount =
            _nullishCoalesce(request.fromWallet.balanceMap.get(request.fromTokenId), () => ( '0'))
          newRequest = {
            ...request,
            nativeAmount: maxAmount,
            quoteFor: 'from'
          }
        } else {
          newRequest = await _swapHelpers.getMaxSwappable.call(void 0, 
            async r => await fetchSwapQuoteInner(r),
            request
          )
        }
      }
      const swapOrder = await fetchSwapQuoteInner(newRequest)
      return await _swapHelpers.makeSwapPluginQuote.call(void 0, swapOrder)
    }
  }
  return out
} exports.makeRangoPlugin = makeRangoPlugin;
